package FluxY_Lib::MS::DataClass_FA;

################################################################################
# This class provides support for fatty acid methyl esthers and the McLafferty
# fragments.
#
# This class inherits basic functionality from the DataClass parent class,
# initializing it with the fatty acid compounds, the atom list and compound
# name. It extends the basic DataClass to add support for derivatized fragments.
# As such it over-rides the GetCarbonInfo() and GetAtomCount() methods, as well
# as implementing the required PrintMassTables(). Two additional helper classes,
# GenerateMassList() and GetMassList() are also implemented.
#
################################################################################

use strict;
use warnings;

use FluxY_Lib::Chemistry::atomInfo; # User Lib - Atomic Masses.
use FluxY_Lib::Math::Math_Utils;    # User Lib - Simple math functions.

use FluxY_Lib::MS::DataClass;       # User Lib - The common parent class.

our @ISA = qw(FluxY_Lib::MS::DataClass);    # Inherits from DataClass


# List of the atoms found in fatty acid compounds and subsequently considerred
# for natural abundance:
my @atomList = ('C','H','O');
my @atomMass = ();

# Each of the following atom counts and masses is an array of the number of each
# corresponding atom relative to the order specified in the atomList,
# i.e. C, H and O.

# Hash of compound data, (first two array elements are required).
# 1. The compound names (fatty acids), in lower case (for standard comparison).
# 2. The atom count for each fatty acid compound.
my %compounds = (
   c14_0 => ["c14:0", [14,28,2]],
   c16_0 => ["c16:0", [16,32,2]],
   c18_0 => ["c18:0", [18,36,2]],
   c18_1 => ["c18:1", [18,34,2]],
   c18_2 => ["c18:2", [18,32,2]],
   c18_3 => ["c18:3", [18,30,2]],
   c20_0 => ["c20:0", [20,40,2]],
   c20_1 => ["c20:1", [20,38,2]],
   c22_0 => ["c22:0",[22,44,2]],
   c22_1 => ["c22:1",[22,42,2]],
   );
   
# A 2D Hash of compound masses, indexed by compunds key and frags keys.
# Each element is the truncated mass value for the given compound/fragment.
my %massList = ();

# The name and atom count for the methylation of the fatty acid.
my $mAgentName = "Methyl Ester";
my @mAgent     = (1,2,0);

# In addition to the molecular ion, each compound may contain McLafferty
# fragments. These fragments are cleaved from the end of the FA chain, and as 
#such have identical compositions for all compounds.
my %frags = (
   m69 => [[4,7,0], [1,4]],   # McLafferty Fragment M69.
   C2  => [[2,4,2], [1,2]],   # McLafferty Fragment C2
   C3  => [[3,5,2], [1,3]],   # McLafferty Fragment C3
   );
   

################################################################################
# Class constructor
#
# Used to instantiate an object for this class.
#
################################################################################
sub new {
   my ($class) = @_;
   
   # Initialize the atom mass vector.
   @atomMass = &GetElementMasses(@atomList);

   &GenerateMassList();
   
   my $self = $class->SUPER::new(\%compounds, \@atomList, "Fatty Acids (FA)");

   bless $self, $class;

   return $self;
}


########################################################
#                                                      #
#   Over-ridden DataClass methods.                     #
#                                                      #
#   - to support the derivating agent and fragments.   #
#                                                      #
########################################################

################################################################################
# Subroutine: GetCarbonInfo
#
# Returns the number of carbon atoms, and the relative positional carbon string
# for the given fragment.
#
################################################################################
sub GetCarbonInfo{
   my ($self, $compound, $fragMass) = @_;
   
   # Full Fatty Acid:
   if ( $fragMass == $massList{$compound}{'M-0'} ) {
      my $numCarbons = $self->GetNumCarbons($compound);
      
      my $fragAtomStr = "1";
      for (my $i=2; $i<=$numCarbons; $i++) { $fragAtomStr .= ",$i"; }
      
      return ($numCarbons, $fragAtomStr);
   }

   # One of the McLafferty Fragments.
   foreach my $frag (sort keys %frags) {
      if ( $fragMass == $massList{'c14_0'}{$frag} ) {
         my ($firstC, $lastC) = @{${$frags{$frag}}[1]};
         
         my $numCarbons = $lastC - $firstC + 1;

         my $fragAtomStr = "";
         while ( $firstC < $lastC ) {
            $fragAtomStr .= "$firstC,";
            $firstC++;
         }         
         $fragAtomStr .= "$lastC";
         
         return ($numCarbons, $fragAtomStr);
      }
   }

   # If the fragment is unidentified, return no values:
   return;
}


################################################################################
# Subroutine: GetAtomCount
#
# Returns the atom count corresponding to the specific fatty acid methyl esther,
# or the McLafferty fragment.
#
################################################################################
sub GetAtomCount{
   my ($self, $compound, $fragMass) = @_;
      
   # Full Fatty Acid:
   if ( $fragMass == $massList{$compound}{'M-0'} ) {
      my @totAtoms = @{${$compounds{$compound}}[1]};
      for(my $i=0; $i<@totAtoms; $i++) {
         $totAtoms[$i] += $mAgent[$i];
      }
      
      return (@totAtoms);
   }

   # One of the McLafferty Fragments.
   foreach my $frag (sort keys %frags) {
      if ( $fragMass == $massList{'c14_0'}{$frag} ) {
         my @totAtoms = @{${$frags{$frag}}[0]};
         for(my $i=0; $i<@totAtoms; $i++) {
            $totAtoms[$i] += $mAgent[$i];
         }

         return (@totAtoms);
      }
   }

   # If the fragment is unidentified, return no values:
   return;
}


######################################
#                                    #
#   Implemented DataClass methods.   #
#                                    #
######################################

################################################################################
# Subroutine: PrintMassTables
#
# Prints a table containing the atom count / fragment mass for each Methylated
# Fatty Acid, and for the McLafferty fragment.
#
################################################################################
sub PrintMassTables{
   my ($self) = @_;

   my $outFile = "FA_Data_Table.txt";

   open (FILE, ">" . $outFile) or FileSystemError(900, $outFile, $!);
   
   print FILE "Fatty Acid $mAgentName Data Table:\n\n\t";
   
   print FILE "\t" . join("\t", @atomList) . "\tMass\tC-Backbone\n";

   # Full Fatty Acid:
   foreach my $compound (sort keys %compounds) {               
      print FILE $compound . "\t";
      
      my @totAtoms   = @{${$compounds{$compound}}[1]};
	   my $numCarbons = $self->GetNumCarbons($compound);

      for(my $i=0; $i<@totAtoms; $i++) {
         $totAtoms[$i] += $mAgent[$i];
      }
      print FILE "\t" . join("\t", @totAtoms) . "\t" . $massList{$compound}{'M-0'} . "\t1-" . $numCarbons . "\n";
   }
   
   print FILE "\nMcLafferty Fragments (the same for each compound):\n\n";

   # One of the McLafferty Fragments.
   foreach my $frag (sort keys %frags) {
      my @fragAtomCount = @{${$frags{$frag}}[0]};
      
      for(my $i=0; $i<@fragAtomCount; $i++) { 
			$fragAtomCount[$i] += $mAgent[$i];
		}
		
      print FILE "\t" . $frag . "\t" . join("\t", @fragAtomCount) . "\t" . $massList{'c14_0'}{$frag} . "\t" . join("-", @{${$frags{$frag}}[1]}). "\n";
   }

   close(FILE)                 or FileSystemError(901, $outFile, $!);
}


#######################
#                     #
#   Public methods.   #
#                     #
#######################

################################################################################
# Subroutine: GenerateMassList
#
# Generates a list of mass values in a hash (indexed by compound key) of hashes
# (indexed by fragment key).
#
################################################################################
sub GenerateMassList {
   my $mass;
   foreach my $compound (sort keys %compounds) {
      my %list = ();

      my @atoms = @{${$compounds{$compound}}[1]};
            
      $mass = 0;      
      for(my $i=0; $i<@atoms; $i++) {
         $mass += ($atoms[$i] + $mAgent[$i]) * $atomMass[$i];
      }
      $list{'M-0'} = &Round($mass);
      
      foreach my $frag (sort keys %frags) { 
         @atoms = @{${$frags{$frag}}[0]};
               
         $mass = 0;      
         for(my $i=0; $i<@atoms; $i++) {
            $mass += ($atoms[$i] + $mAgent[$i]) * $atomMass[$i];
         }
         
         $list{$frag} = &Round($mass);
      }
      
      $massList{$compound} = \%list;
   }
   
}


################################################################################
# Subroutine: GetMassList
#
# Return the massList hash.
#
################################################################################
sub GetMassList {
   return %massList;
}

1;